package java.beans;

import java.lang.reflect.Array;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.IdentityHashMap;
import java.nio.charset.Charset;
import java.io.OutputStream;
import java.util.List;
import java.util.Map;
import java.io.OutputStreamWriter;
import java.nio.charset.CharsetEncoder;

public class XMLEncoder extends Encoder implements AutoCloseable
{
    private final CharsetEncoder encoder;
    private final String charset;
    private final boolean declaration;
    private OutputStreamWriter out;
    private Object owner;
    private int indentation;
    private boolean internal;
    private Map<Object, XMLEncoder.ValueData> valueToExpression;
    private Map<Object, List<Statement>> targetToStatementList;
    private boolean preambleWritten;
    private NameGenerator nameGenerator;
    
    public XMLEncoder(OutputStream outputStream) {
        this(outputStream, "UTF-8", true, 0);
    }
    
    public XMLEncoder(OutputStream out, String s, boolean declaration, int indentation) {
        this.indentation = 0;
        this.internal = false;
        this.preambleWritten = false;
        if (out == null) {
            throw new IllegalArgumentException("the output stream cannot be null");
        }
        if (indentation < 0) {
            throw new IllegalArgumentException("the indentation must be >= 0");
        }
        Charset forName = Charset.forName(s);
        this.encoder = forName.newEncoder();
        this.charset = s;
        this.declaration = declaration;
        this.indentation = indentation;
        this.out = new OutputStreamWriter(out, forName.newEncoder());
        this.valueToExpression = new IdentityHashMap();
        this.targetToStatementList = new IdentityHashMap();
        this.nameGenerator = new NameGenerator();
        return;
    }
    
    public void setOwner(Object owner) {
        this.owner = owner;
        this.writeExpression(new Expression(this, "getOwner", new Object[0]));
    }
    
    public Object getOwner() {
        return this.owner;
    }
    
    public void writeObject(Object o) {
        if (this.internal) {
            super.writeObject(o);
        }
        else {
            this.writeStatement(new Statement(this, "writeObject", new Object[] { o }));
        }
    }
    
    private List<Statement> statementList(Object o) {
        List<Object> list = (List)this.targetToStatementList.get(o);
        if (list == null) {
            list = new ArrayList();
            this.targetToStatementList.put(o, list);
            return list;
        }
        return list;
    }
    
    private void mark(Object o, boolean b) {
        if (o == null || o == this) {
            return;
        }
        XMLEncoder.ValueData valueData = this.getValueData(o);
        Expression exp = valueData.exp;
        if (o.getClass() == String.class && exp == null) {
            return;
        }
        if (b) {
            ++valueData.refs;
        }
        if (valueData.marked) {
            return;
        }
        valueData.marked = true;
        Object target = exp.getTarget();
        this.mark(exp);
        if (!(target instanceof Class)) {
            this.statementList(target).add(exp);
            ++valueData.refs;
        }
        return;
    }
    
    private void mark(Statement statement) {
        Object[] arguments = statement.getArguments();
        for (int i = 0; i < arguments.length; ++i) {
            this.mark(arguments[i], true);
        }
        this.mark(statement.getTarget(), statement instanceof Expression);
    }
    
    @Override
    public void writeStatement(Statement obj) {
        int internal = this.internal;
        this.internal = true;
        try {
            super.writeStatement(obj);
            this.mark(obj);
            Object target = obj.getTarget();
            if ((target instanceof Field)) {
                String methodName = obj.getMethodName();
                Object[] arguments = obj.getArguments();
                if (methodName != null && arguments != null) {
                    if (methodName.equals("get") && arguments.length == 1) {
                        target = arguments[0];
                    }
                    else if (methodName.equals("set") && arguments.length == 2) {
                        target = arguments[0];
                    }
                }
            }
            this.statementList(target).add(obj);
        }
        catch (Exception cause) {
            this.getExceptionListener().exceptionThrown(new Exception("XMLEncoder: discarding statement " + obj, cause));
        }
        this.internal = internal;
    }
    
    @Override
    public void writeExpression(Expression exp) {
        int internal = this.internal;
        this.internal = true;
        Object value = this.getValue(exp);
        if (this.get(value) == null || (value instanceof String && internal == 0)) {
            super.writeExpression(this.getValueData(value).exp = exp);
        }
        this.internal = internal;
    }
    
    public void flush() {
        if (!this.preambleWritten) {
            if (this.declaration) {
                this.writeln("<?xml version=" + this.quote("1.0") + " encoding=" + this.quote(this.charset) + "?>");
            }
            this.writeln("<java version=" + this.quote(System.getProperty("java.version")) + " class=" + this.quote(XMLDecoder.class.getName()) + ">");
            this.preambleWritten = true;
        }
        ++this.indentation;
        List<Statement> statementList = this.statementList(this);
        while (!statementList.isEmpty()) {
            Statement statement = (Statement)statementList.remove(0);
            if ("writeObject".equals(statement.getMethodName())) {
                this.outputValue(statement.getArguments()[0], this, true);
            }
            else {
                this.outputStatement(statement, this, false);
            }
        }
        --this.indentation;
        for (Statement statement2 = this.getMissedStatement(); statement2 != null; statement2 = this.getMissedStatement()) {
            this.outputStatement(statement2, this, false);
        }
        try {
            this.out.flush();
        }
        catch (IOException ex) {
            this.getExceptionListener().exceptionThrown(ex);
        }
        this.clear();
    }
    
    @Override
    void clear() {
        super.clear();
        this.nameGenerator.clear();
        this.valueToExpression.clear();
        this.targetToStatementList.clear();
    }
    
    Statement getMissedStatement() {
        for (List list : this.targetToStatementList.values()) {
            int i = 0;
            while (i < list.size()) {
                if (Statement.class == ((Object)list.get(i)).getClass()) {
                    return (Statement)list.remove(i);
                }
                ++i;
            }
        }
        return null;
    }
    
    @Override
    public void close() {
        this.flush();
        this.writeln("</java>");
        try {
            this.out.close();
        }
        catch (IOException ex) {
            this.getExceptionListener().exceptionThrown(ex);
        }
    }
    
    private String quote(String str) {
        return "\"" + str + "\"";
    }
    
    private XMLEncoder.ValueData getValueData(Object o) {
        XMLEncoder.ValueData valueData = (XMLEncoder.ValueData)this.valueToExpression.get(o);
        if (valueData == null) {
            valueData = new XMLEncoder.ValueData();
            this.valueToExpression.put(o, valueData);
            return valueData;
        }
        return valueData;
    }
    
    private static boolean isValidCharCode(int n) {
        return (32 <= n && n <= 55295) || 10 == n || 9 == n || 13 == n || (57344 <= n && n <= 65533) || (65536 <= n && n <= 1114111);
    }
    
    private void writeln(String str) {
        try {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < this.indentation; ++i) {
                sb.append(' ');
            }
            sb.append(str);
            sb.append('\n');
            this.out.write(sb.toString());
        }
        catch (IOException ex) {
            this.getExceptionListener().exceptionThrown(ex);
        }
    }
    
    private void outputValue(Object obj, Object o, boolean b) {
        if (obj == null) {
            this.writeln("<null/>");
            return;
        }
        if ((obj instanceof Class)) {
            this.writeln("<class>" + ((Class)obj).getName() + "</class>");
            return;
        }
        XMLEncoder.ValueData valueData = this.getValueData(obj);
        if (valueData.exp != null) {
            Object target = valueData.exp.getTarget();
            String methodName = valueData.exp.getMethodName();
            if (target == null || methodName == null) {
                throw new NullPointerException(((target == null) ? "target" : "methodName") + " should not be null");
            }
            if (b && target instanceof Field && methodName.equals("get")) {
                Field field = (Field)target;
                if (Modifier.isStatic(field.getModifiers())) {
                    this.writeln("<object class=" + this.quote(field.getDeclaringClass().getName()) + " field=" + this.quote(field.getName()) + "/>");
                    return;
                }
            }
            Class<Object> primitiveType = primitiveTypeFor(obj.getClass());
            if (primitiveType != null && target == obj.getClass() && methodName.equals("new")) {
                String name = primitiveType.getName();
                if (primitiveType == Character.TYPE) {
                    int charValue = (char)obj;
                    if (!isValidCharCode(charValue)) {
                        this.writeln(createString(charValue));
                        return;
                    }
                    obj = quoteCharCode(charValue);
                    if (obj == null) {
                        obj = Character.valueOf((char)charValue);
                    }
                }
                this.writeln("<" + name + ">" + obj + "</" + name + ">");
                return;
            }
        }
        else if ((obj instanceof String)) {
            this.writeln(this.createString((String)obj));
            return;
        }
        if (valueData.name != null) {
            if (b) {
                this.writeln("<object idref=" + this.quote(valueData.name) + "/>");
            }
            else {
                this.outputXML("void", " idref=" + this.quote(valueData.name), obj, new Object[0]);
            }
        }
        else if (valueData.exp != null) {
            this.outputStatement(valueData.exp, o, b);
        }
        return;
    }
    
    private static String quoteCharCode(int n) {
        switch (n) {
            case 38: {
                return "&amp;";
            }
            case 60: {
                return "&lt;";
            }
            case 62: {
                return "&gt;";
            }
            case 34: {
                return "&quot;";
            }
            case 39: {
                return "&apos;";
            }
            case 13: {
                return "&#13;";
            }
            default: {
                return null;
            }
        }
    }
    
    private static String createString(int i) {
        return "<char code=\"#" + Integer.toString(i, 16) + "\"/>";
    }
    
    private String createString(String s) {
        StringBuilder sb = new StringBuilder();
        sb.append("<string>");
        int i = 0;
        while (i < s.length()) {
            int codePoint = s.codePointAt(i);
            int charCount = Character.charCount(codePoint);
            if (isValidCharCode(codePoint) && this.encoder.canEncode(s.substring(i, i + charCount))) {
                String quoteCharCode = quoteCharCode(codePoint);
                if (quoteCharCode != null) {
                    sb.append(quoteCharCode);
                }
                else {
                    sb.appendCodePoint(codePoint);
                }
                i += charCount;
            }
            else {
                sb.append(createString((int)s.charAt(i)));
                ++i;
            }
        }
        sb.append("</string>");
        return sb.toString();
    }
    
    private void outputStatement(Statement statement, Object o, boolean b) {
        Object target = statement.getTarget();
        String methodName = statement.getMethodName();
        if (target == null || methodName == null) {
            throw new NullPointerException(((target == null) ? "target" : "methodName") + " should not be null");
        }
        Object[] arguments = statement.getArguments();
        int n = statement.getClass() == Expression.class;
        Object o2 = n ? this.getValue((Expression)statement) : null;
        String s = (n && b) ? "object" : "void";
        String s2 = "";
        XMLEncoder.ValueData valueData = this.getValueData(o2);
        if (target != o) {
            if (target == Array.class && methodName.equals("newInstance")) {
                s = "array";
                s2 = s2 + " class=" + this.quote(((Class)arguments[0]).getName()) + " length=" + this.quote(arguments[1].toString());
                arguments = new Object[0];
            }
            else if (target.getClass() == Class.class) {
                s2 = s2 + " class=" + this.quote(((Class<Array>)target).getName());
            }
            else {
                valueData.refs = 2;
                if (valueData.name == null) {
                    XMLEncoder.ValueData valueData2 = this.getValueData(target);
                    ++valueData2.refs;
                    List<Statement> statementList = this.statementList(target);
                    if (!statementList.contains(statement)) {
                        statementList.add(statement);
                    }
                    this.outputValue(target, o, false);
                }
                if (!(n == 0)) {
                    this.outputValue(o2, o, b);
                }
                return;
            }
        }
        if (n && valueData.refs > 1) {
            String instanceName = this.nameGenerator.instanceName(o2);
            valueData.name = instanceName;
            s2 = s2 + " id=" + this.quote(instanceName);
        }
        if ((n == 0 && methodName.equals("set") && arguments.length == 2 && arguments[0] instanceof Integer) || (n && methodName.equals("get") && arguments.length == 1 && arguments[0] instanceof Integer)) {
            s2 = s2 + " index=" + this.quote(arguments[0].toString());
            arguments = ((arguments.length == 1) ? new Object[0] : new Object[] { arguments[1] });
        }
        else if ((n == 0 && methodName.startsWith("set") && arguments.length == 1) || (n && methodName.startsWith("get") && arguments.length == 0)) {
            if (3 < methodName.length()) {
                s2 = s2 + " property=" + this.quote(Introspector.decapitalize(methodName.substring(3)));
            }
        }
        else if (!methodName.equals("new") && !methodName.equals("newInstance")) {
            s2 = s2 + " method=" + this.quote(methodName);
        }
        this.outputXML(s, s2, o2, arguments);
        return;
    }
    
    private void outputXML(String str, String s, Object o, Object... array) {
        List<Statement> statementList = this.statementList(o);
        if (array.length == 0 && statementList.size() == 0) {
            this.writeln("<" + str + s + "/>");
            return;
        }
        this.writeln("<" + str + s + ">");
        ++this.indentation;
        for (int i = 0; i < array.length; ++i) {
            this.outputValue(array[i], null, true);
        }
        while (!statementList.isEmpty()) {
            this.outputStatement((Statement)statementList.remove(0), o, false);
        }
        --this.indentation;
        this.writeln("</" + str + ">");
        return;
    }
    
    static Class primitiveTypeFor(Class clazz) {
        if (clazz == Boolean.class) {
            return Boolean.TYPE;
        }
        if (clazz == Byte.class) {
            return Byte.TYPE;
        }
        if (clazz == Character.class) {
            return Character.TYPE;
        }
        if (clazz == Short.class) {
            return Short.TYPE;
        }
        if (clazz == Integer.class) {
            return Integer.TYPE;
        }
        if (clazz == Long.class) {
            return Long.TYPE;
        }
        if (clazz == Float.class) {
            return Float.TYPE;
        }
        if (clazz == Double.class) {
            return Double.TYPE;
        }
        if (clazz == Void.class) {
            return Void.TYPE;
        }
        return null;
    }
    
    private class ValueData
    {
        public int refs;
        public boolean marked;
        public String name;
        public Expression exp;
        
        private ValueData() {
            this.refs = 0;
            this.marked = false;
            this.name = null;
            this.exp = null;
        }
    }
}
